/*
 *   files.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"								// I/O register definitions
#include "kml.h"
#include "debugger.h"

TCHAR  szEmuDirectory[MAX_PATH];
TCHAR  szCurrentKml[MAX_PATH];
TCHAR  szBackupKml[MAX_PATH];
TCHAR  szCurrentFilename[MAX_PATH];
TCHAR  szBackupFilename[MAX_PATH];
TCHAR  szBufferFilename[MAX_PATH];

LPBYTE pbyRom = NULL;
DWORD  dwRomSize = 0;
BYTE   cCurrentRomType = 0;					// Model -> hardware
UINT   nCurrentClass = 0;					// Class -> derivate

// document signatures
static BYTE pbySignature[] = "Emu42 Document\xFE";
static HANDLE hCurrentFile = NULL;

static CHIPSET_M BackupChipset;
static CHIPSET_S BackupChipsetS;

BOOL   bBackup = FALSE;

//################
//#
//#    Subroutine for Write on Stack
//#
//################

WORD WriteStack(LPBYTE lpBuf,DWORD dwSize)	// separated from LoadObject()
{
	BOOL   bBinary;
	DWORD  dwAddress, i;

	LPBYTE lpHead = BINARYHEADER28;

	bBinary =  ((lpBuf[dwSize+0]==lpHead[0])
	        &&  (lpBuf[dwSize+1]==lpHead[1])
	        &&  (lpBuf[dwSize+2]==lpHead[2])
	        &&  (lpBuf[dwSize+3]==lpHead[3])
	        &&  (lpBuf[dwSize+4]==lpHead[4])
	        &&  (lpBuf[dwSize+5]==lpHead[5])
	        &&  (lpBuf[dwSize+6]==lpHead[6]));

	for (i = 0; i < dwSize; i++)
	{
		BYTE byTwoNibs = lpBuf[i+dwSize];
		lpBuf[i*2  ] = (BYTE)(byTwoNibs&0xF);
		lpBuf[i*2+1] = (BYTE)(byTwoNibs>>4);
	}

	if (bBinary == TRUE)
	{ // load as binary
		dwSize    = RPL_ObjectSize(lpBuf+16);
		dwAddress = RPL_CreateTemp(dwSize);
		if (dwAddress == 0) return S_ERR_BINARY;

		Nwrite(lpBuf+16,dwAddress,dwSize);
	}
	else
	{ // load as string
		BYTE lpHead[5];
		dwSize *= 2;
		dwAddress = RPL_CreateTemp(dwSize+10);
		if (dwAddress == 0) return S_ERR_ASCII;
		Nunpack(lpHead,0x02A4E,5);			// String
		Nwrite(lpHead,dwAddress,5);
		Nunpack(lpHead,dwSize+5,5);			// length of String
		Nwrite(lpHead,dwAddress+5,5);
		Nwrite(lpBuf,dwAddress+10,dwSize);	// data
	}
	RPL_Push(dwAddress);
	return S_ERR_NO;
}



//################
//#
//#    Window Position Tools
//#
//################

VOID SetWindowLocation(HWND hWnd,INT nPosX,INT nPosY)
{
	WINDOWPLACEMENT wndpl;
	RECT *pRc = &wndpl.rcNormalPosition;

	wndpl.length = sizeof(wndpl);
	GetWindowPlacement(hWnd,&wndpl);
	pRc->right = pRc->right - pRc->left + nPosX;
	pRc->bottom = pRc->bottom - pRc->top + nPosY;
	pRc->left = nPosX;
	pRc->top = nPosY;
	SetWindowPlacement(hWnd,&wndpl);
	return;
}



//################
//#
//#    Directory Helper Tool
//#
//################

LPCTSTR FullFilename(LPCTSTR lpszFilename)
{
	static TCHAR szFullname[MAX_PATH];

	INT nPos;

	// copy directory path
	lstrcpyn(szFullname,szEmuDirectory,ARRAYSIZEOF(szFullname));
	nPos = lstrlen(szFullname);				// end of path name

	if (nPos > 0 && nPos < ARRAYSIZEOF(szFullname) && szFullname[nPos-1] != _T('\\'))
	{
		szFullname[nPos++] = _T('\\');
	}

	lstrcpyn(&szFullname[nPos],lpszFilename,ARRAYSIZEOF(szFullname)-nPos);
	return szFullname;
}



//################
//#
//#    Patch
//#
//################

// Lewis CRC Pass 0
#define a0	0x1C							// Start Address
#define d0	(n0*16)							// Address offset
#define n0	0x555							// Reads/Half-Sector
#define s0	3								// #Sectors (Sector Size=2*d)

// Lewis CRC Pass 1
#define a1	0x0								// Start Address
#define d1	(n1*16)							// Address offset
#define n1	1								// #Reads/Half-Sector
#define s1	0x1000							// #Sectors (Sector Size=2*d)

// rebuild of the calculator =CHECKSUM function
static WORD Checksum(LPBYTE pbyROM, DWORD dwStart, DWORD dwOffset, INT nReads, INT nSector)
{
	int i,j;

	WORD wCrc = 0;

	for (;nSector > 0; --nSector)			// evalute each sector
	{
		DWORD dwAddr1 = dwStart;
		DWORD dwAddr2 = dwStart + dwOffset;

		for (i = 0; i < nReads; ++i)		// no. of reads in sector
		{
			for (j = 0; j < 16; ++j) wCrc = UpCRC(wCrc,pbyROM[dwAddr1++]);
			for (j = 0; j < 16; ++j) wCrc = UpCRC(wCrc,pbyROM[dwAddr2++]);
		}

		dwStart += 2 * dwOffset;			// next start page
	}
	return wCrc;
}

static VOID CorrectCrc(BYTE *a, WORD wCrc)
{
	INT s;

	wCrc = ~wCrc;							// correction word to 0xFFFF

	// unpack and write CRC
	for (s = 3; s >= 0; --s) { a[s] = (BYTE)(wCrc&0xf); wCrc>>=4; }
}

static VOID RebuildRomCrc(VOID)
{
	// patch master Lewis chip
	ZeroMemory(&pbyRom[0x1FFF8],8);			// delete old Crc's
	CorrectCrc(&pbyRom[0x1FFF8],Checksum(pbyRom,a0,d0,n0,s0));
	CorrectCrc(&pbyRom[0x1FFFC],Checksum(pbyRom,a1,d1,n1,s1));

	if (dwRomSize == _KB(96))				// with external ROM
	{
		ZeroMemory(&pbyRom[0x2FFFC],4);		// delete old Crc
		CorrectCrc(&pbyRom[0x2FFFC],Checksum(pbyRom,a1+_KB(64),d1,n1,s1/2));
	}

	if (dwRomSize == _KB(128))				// with slave Lewis chip
	{
		ZeroMemory(&pbyRom[0x3FFF8],8);		// delete old Crc's
		CorrectCrc(&pbyRom[0x3FFF8],Checksum(pbyRom,a0+_KB(64),d0,n0,s0));
		CorrectCrc(&pbyRom[0x3FFFC],Checksum(pbyRom,a1+_KB(64),d1,n1,s1));
	}
	return;
}

static __inline BYTE Asc2Nib(BYTE c)
{
	if (c<'0') return 0;
	if (c<='9') return c-'0';
	if (c<'A') return 0;
	if (c<='F') return c-'A'+10;
	if (c<'a') return 0;
	if (c<='f') return c-'a'+10;
	return 0;
}

BOOL PatchRom(LPCTSTR szFilename)
{
	HANDLE hFile = NULL;
	DWORD  dwFileSizeLow = 0;
	DWORD  dwFileSizeHigh = 0;
	DWORD  lBytesRead = 0;
	LPBYTE lpStop,lpBuf = NULL;
	DWORD  dwAddress = 0;
	UINT   nPos = 0;

	if (pbyRom == NULL) return FALSE;
	hFile = CreateFile(FullFilename(szFilename), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return FALSE;
	dwFileSizeLow = GetFileSize(hFile, &dwFileSizeHigh);
	if (dwFileSizeLow <= 5)
	{ // file is too small.
		CloseHandle(hFile);
		return FALSE;
	}
	if (dwFileSizeHigh != 0)
	{ // file is too large.
		CloseHandle(hFile);
		return FALSE;
	}
	lpBuf = HeapAlloc(hHeap,0,dwFileSizeLow+1);
	if (lpBuf == NULL)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	ReadFile(hFile, lpBuf, dwFileSizeLow, &lBytesRead, NULL);
	CloseHandle(hFile);
	lpBuf[dwFileSizeLow] = 0;
	nPos = 0;
	while (lpBuf[nPos])
	{
		do // remove blank space
		{
			if (  (lpBuf[nPos]!=' ')
				&&(lpBuf[nPos]!='\n')
				&&(lpBuf[nPos]!='\r')
				&&(lpBuf[nPos]!='\t')) break;
			nPos++;
		} while (lpBuf[nPos]);
		if (lpBuf[nPos]==';') // comment ?
		{
			do
			{
				nPos++;
				if (lpBuf[nPos]=='\n')
				{
					nPos++;
					break;
				}
			} while (lpBuf[nPos]);
			continue;
		}
		dwAddress = strtoul(&lpBuf[nPos],(CHAR **)&lpStop,16);
		nPos += lpStop-&lpBuf[nPos]+1;
		if (*lpStop != ':' || *lpStop == 0)
			continue;
		while (lpBuf[nPos])
		{
			if (isxdigit(lpBuf[nPos]) == FALSE) break;
			// patch ROM
			pbyRom[dwAddress] = Asc2Nib(lpBuf[nPos]);
			dwAddress = (dwAddress+1) % dwRomSize;
			++nPos;
		}
	}
	HeapFree(hHeap,0,lpBuf);
	RebuildRomCrc();						// rebuild the ROM CRC's
	return TRUE;
}



//################
//#
//#    ROM
//#
//################

static WORD CrcRom(VOID)					// calculate fingerprint of ROM
{
	DWORD dwCount;
	DWORD dwCrc = 0;

	_ASSERT(pbyRom);						// view on ROM
	// use checksum, because it's faster
	for (dwCount = 0;dwCount < dwRomSize; dwCount+=sizeof(dwCrc))
		dwCrc += *((DWORD *) &pbyRom[dwCount]);

	return (WORD) dwCrc;
}

BOOL MapRom(LPCTSTR szFilename)
{
	HANDLE hRomFile;
	DWORD  dwSize,dwFileSize,dwRead;
	INT    i;

	if (pbyRom != NULL)
	{
		return FALSE;
	}
	hRomFile = CreateFile(FullFilename(szFilename),
						  GENERIC_READ,
						  FILE_SHARE_READ,
						  NULL,
						  OPEN_EXISTING,
						  FILE_FLAG_SEQUENTIAL_SCAN,
						  NULL);
	if (hRomFile == INVALID_HANDLE_VALUE)
	{
		hRomFile = NULL;
		return FALSE;
	}
	dwFileSize = GetFileSize(hRomFile, NULL);
	if (dwFileSize <= 4)
	{ // file is too small.
		CloseHandle(hRomFile);
		hRomFile = NULL;
		dwRomSize = 0;
		return FALSE;
	}

	// read the first 4 bytes
	ReadFile(hRomFile,&dwSize,sizeof(dwSize),&dwRead,NULL);

	dwRomSize = dwFileSize;					// calculate ROM image buffer size
	if ((dwSize & 0xF0F0F0F0) != 0)			// packed ROM image ->
		dwRomSize *= 2;						// unpacked ROM image has double size

	pbyRom = VirtualAlloc(NULL,dwRomSize,MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
	if (pbyRom == NULL)
	{
		CloseHandle(hRomFile);
		hRomFile = NULL;
		dwRomSize = 0;
		return FALSE;
	}

	*(DWORD *) pbyRom = dwSize;				// save first 4 bytes

	// load rest of file content
	ReadFile(hRomFile,&pbyRom[sizeof(dwSize)],dwFileSize - sizeof(dwSize),&dwRead,NULL);
	_ASSERT(dwFileSize - sizeof(dwSize) == dwRead);
	CloseHandle(hRomFile);

	if (dwRomSize != dwFileSize)			// packed ROM image
	{
		i = dwFileSize - 1;					// source start address
		dwSize = dwRomSize - 1;				// destination start address
		while (i >= 0)						// unpack source
		{
			BYTE byValue = pbyRom[i--];
			pbyRom[dwSize--] = byValue >> 4;
			pbyRom[dwSize--] = byValue & 0xF;
		}
	}
	return TRUE;
}

VOID UnmapRom(VOID)
{
	if (pbyRom==NULL) return;
	VirtualFree(pbyRom,0,MEM_RELEASE);
	pbyRom = NULL;
	dwRomSize = 0;
	return;
}



//################
//#
//#    Documents
//#
//################

VOID ResetDocument(VOID)
{
	if (szCurrentKml[0])
	{
		KillKML();
	}
	if (hCurrentFile)
	{
		CloseHandle(hCurrentFile);
		hCurrentFile = NULL;
	}
	szCurrentKml[0] = 0;
	szCurrentFilename[0]=0;

	if (Chipset.NCE2)  HeapFree(hHeap,0,Chipset.NCE2);
	if (Chipset.NCE3)  HeapFree(hHeap,0,Chipset.NCE3);
	if (ChipsetS.NCE2) HeapFree(hHeap,0,ChipsetS.NCE2);
	if (ChipsetS.NCE3) HeapFree(hHeap,0,ChipsetS.NCE3);
	ZeroMemory(&Chipset,sizeof(Chipset));
	ZeroMemory(&ChipsetS,sizeof(ChipsetS));
	ZeroMemory(&RMap,sizeof(RMap));			// delete MMU mappings
	ZeroMemory(&WMap,sizeof(WMap));
	UpdateWindowStatus();
	return;
}

BOOL NewDocument(VOID)
{
	ResetDocument();

	if (!DisplayChooseKml(0)) goto restore;
	if (!InitKML(szCurrentKml,FALSE)) goto restore;
	Chipset.type = cCurrentRomType;

	// MMU mappings default settings for Pioneer series
	Chipset.RegBase  = 0x40300;				// MMU Registers (16 bits)
	Chipset.RegMask  = 0xFFFF0;				// n-island mask
	Chipset.DispBase = 0x40000;				// MMU Display/Timer (10 bits)
	Chipset.DispMask = 0xFFC00;				// n-island mask
	Chipset.NCE2Base = 0x50000;				// MMU NCE2 Memory (10 bits)
	Chipset.NCE2Mask = 0xF0000;				// n-island mask
	Chipset.NCE3Base = 0x20000;				// MMU NCE3 Memory (10 bits)
	Chipset.NCE3Mask = 0xE0000;				// n-island mask
	Chipset.RomBase  = 0x00000;				// MMU ROM (6 bits)
	Chipset.RomMask  = 0xE0000;				// n-island mask

	Chipset.RomSize  = dwRomSize;			// size of ROM

	if (Chipset.RomSize > _KB(64))			// external ROM
	{
		Chipset.RomSize  = _KB(64);			// size of mask ROM

		Chipset.NCE3Ram  = FALSE;			// ROM
		Chipset.NCE3Size = dwRomSize-Chipset.RomSize;
	}

	Chipset.NCE2Ram  = TRUE;				// RAM
	// physical DRAM chip (default 8KB or size of KML Class argument)
	Chipset.NCE2Size = _KB((nCurrentClass == 0) ? 8 : nCurrentClass);

	Chipset.IORam[RATECTL] = 0x7;			// CPU speed (mask-programmed)
	Chipset.IORam[LPE] = RST;				// set ReSeT bit at power up

	if (Chipset.type == 'O')				// Orlando, HP28S
	{
		Chipset.bSlave = TRUE;

		Chipset.RegBase   = 0xFFF00;		// MMU Registers (16 bits)
		Chipset.RegMask   = 0xFFFF0;		// n-island mask
		Chipset.DispBase  = 0xFFC00;		// MMU Display/Timer (10 bits)
		Chipset.DispMask  = 0xFFC00;		// n-island mask
		Chipset.NCE2Base  = 0xC0000;		// MMU NCE2 Memory (10 bits)
		Chipset.NCE2Mask  = 0xE0000;		// n-island mask
		Chipset.NCE3Base  = 0x80000;		// MMU NCE3 Memory (10 bits)
		Chipset.NCE3Mask  = 0xC0000;		// n-island mask
		Chipset.RomBase   = 0x00000;		// MMU ROM (6 bits)
		Chipset.RomMask   = 0xE0000;		// n-island mask

		ChipsetS.RegBase  = 0xFFB00;		// MMU Registers (16 bits)
		ChipsetS.RegMask  = 0xFFFF0;		// n-island mask
		ChipsetS.DispBase = 0xFF800;		// MMU Display/Timer (10 bits)
		ChipsetS.DispMask = 0xFFC00;		// n-island mask
		ChipsetS.NCE2Base = 0x00000;		// MMU NCE2 Memory (10 bits)
		ChipsetS.NCE2Mask = 0x00000;		// n-island mask (unused)
		ChipsetS.NCE3Base = 0x00000;		// MMU NCE3 Memory (10 bits)
		ChipsetS.NCE3Mask = 0x00000;		// n-island mask (unused)
		ChipsetS.RomBase  = 0x20000;		// MMU ROM (6 bits)
		ChipsetS.RomMask  = 0xE0000;		// n-island mask

		Chipset.RomSize   = _KB(64);		// size of master mask ROM
		ChipsetS.RomSize  = _KB(64);		// size of slave  mask ROM

		Chipset.NCE2Ram   = TRUE;			// RAM
		Chipset.NCE2Size  = _KB(32);		// physical DRAM chip

		Chipset.NCE3Ram   = FALSE;			// open data bus
		Chipset.NCE3Size  = 0;

		ChipsetS.IORam[RATECTL] = 0x7;		// CPU speed (mask-programmed)
		ChipsetS.IORam[LPE] = RST;			// set ReSeT bit at power up
	}

	if (Chipset.type == 'Y')				// Tycoon II, HP19BII
	{
		Chipset.bSlave = TRUE;

		Chipset.RegBase   = 0xFFF00;		// MMU Registers (16 bits)
		Chipset.RegMask   = 0xFFFF0;		// n-island mask
		Chipset.DispBase  = 0xFFC00;		// MMU Display/Timer (10 bits)
		Chipset.DispMask  = 0xFFC00;		// n-island mask
		Chipset.NCE2Base  = 0xC4000;		// MMU NCE2 Memory (10 bits)
		Chipset.NCE2Mask  = 0xFC000;		// n-island mask
		Chipset.NCE3Base  = 0x00000;		// MMU NCE3 Memory (10 bits)
		Chipset.NCE3Mask  = 0x00000;		// n-island mask (unused)
		Chipset.RomBase   = 0x00000;		// MMU ROM (6 bits)
		Chipset.RomMask   = 0xE0000;		// n-island mask

		ChipsetS.RegBase  = 0xFFB00;		// MMU Registers (16 bits)
		ChipsetS.RegMask  = 0xFFFF0;		// n-island mask
		ChipsetS.DispBase = 0xFF800;		// MMU Display/Timer (10 bits)
		ChipsetS.DispMask = 0xFFC00;		// n-island mask
		ChipsetS.NCE2Base = 0x00000;		// MMU NCE2 Memory (10 bits)
		ChipsetS.NCE2Mask = 0x00000;		// n-island mask (unused)
		ChipsetS.NCE3Base = 0x00000;		// MMU NCE3 Memory (10 bits)
		ChipsetS.NCE3Mask = 0x00000;		// n-island mask (unused)
		ChipsetS.RomBase  = 0x20000;		// MMU ROM (6 bits)
		ChipsetS.RomMask  = 0xE0000;		// n-island mask

		Chipset.RomSize   = _KB(64);		// size of master mask ROM
		ChipsetS.RomSize  = _KB(64);		// size of slave  mask ROM

		Chipset.NCE2Ram   = TRUE;			// RAM
		Chipset.NCE2Size  = _KB(8);			// physical DRAM chip

		Chipset.NCE3Ram   = FALSE;			// unused
		Chipset.NCE3Size  = 0;

		ChipsetS.IORam[RATECTL] = 0x7;		// CPU speed (mask-programmed)
		ChipsetS.IORam[LPE] = RST;			// set ReSeT bit at power up
	}

	// check MMU controller boundary adjustments
	_ASSERT(Chipset.RegBase   == (Chipset.RegBase   & Chipset.RegMask));
	_ASSERT(Chipset.DispBase  == (Chipset.DispBase  & Chipset.DispMask));
	_ASSERT(Chipset.NCE2Base  == (Chipset.NCE2Base  & Chipset.NCE2Mask));
	_ASSERT(Chipset.NCE3Base  == (Chipset.NCE3Base  & Chipset.NCE3Mask));
	_ASSERT(Chipset.RomBase   == (Chipset.RomBase   & Chipset.RomMask));
	_ASSERT(ChipsetS.RegBase  == (ChipsetS.RegBase  & ChipsetS.RegMask));
	_ASSERT(ChipsetS.DispBase == (ChipsetS.DispBase & ChipsetS.DispMask));
	_ASSERT(ChipsetS.NCE2Base == (ChipsetS.NCE2Base & ChipsetS.NCE2Mask));
	_ASSERT(ChipsetS.NCE3Base == (ChipsetS.NCE3Base & ChipsetS.NCE3Mask));
	_ASSERT(ChipsetS.RomBase  == (ChipsetS.RomBase  & ChipsetS.RomMask));

	// allocate port memory
	if (Chipset.NCE2Ram)
	{
		Chipset.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE2Size);
		_ASSERT(Chipset.NCE2 != NULL);
	}
	if (Chipset.NCE3Ram)
	{
		Chipset.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE3Size);
		_ASSERT(Chipset.NCE3 != NULL);
	}
	if (ChipsetS.NCE2Ram)
	{
		ChipsetS.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE2Size);
		_ASSERT(ChipsetS.NCE2 != NULL);
	}
	if (ChipsetS.NCE3Ram)
	{
		ChipsetS.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE3Size);
		_ASSERT(ChipsetS.NCE3 != NULL);
	}
	LoadBreakpointList(NULL);				// clear debugger breakpoint list
	SaveBackup();
	return TRUE;
restore:
	RestoreBackup();
	ResetBackup();
	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);
	return FALSE;
}

BOOL OpenDocument(LPCTSTR szFilename)
{
	HANDLE  hFile = INVALID_HANDLE_VALUE;
	DWORD   lBytesRead,lSizeofChipset;
	BYTE    pbyFileSignature[sizeof(pbySignature)];
	UINT    ctBytesCompared;
	UINT    nLength;

	SaveBackup();
	ResetDocument();

	// Open file
	if (lstrcmpi(szBackupFilename, szFilename)==0)
	{
		if (YesNoMessage(_T("Do you want to reload this document ?")) == IDNO)
			goto restore;
	}
	hFile = CreateFile(szFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		AbortMessage(_T("This file is missing or already loaded in another instance of Emu42."));
		goto restore;
	}

	// Read and Compare signature
	ReadFile(hFile, pbyFileSignature, sizeof(pbyFileSignature), &lBytesRead, NULL);
	for (ctBytesCompared=0; ctBytesCompared<sizeof(pbyFileSignature); ctBytesCompared++)
	{
		if (pbyFileSignature[ctBytesCompared]!=pbySignature[ctBytesCompared])
		{
			AbortMessage(_T("This file is not a valid Emu42 document."));
			goto restore;
		}
	}

	switch (pbyFileSignature[14])
	{
	case 0xFE: // Emu42 1.0 format
		ReadFile(hFile,&nLength,sizeof(nLength),&lBytesRead,NULL);
		#if defined _UNICODE
		{
			LPSTR szTmp = HeapAlloc(hHeap,0,nLength);
			if (szTmp == NULL)
			{
				AbortMessage(_T("Memory Allocation Failure."));
				goto restore;
			}
			ReadFile(hFile, szTmp, nLength, &lBytesRead, NULL);
			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szTmp, lBytesRead,
								szCurrentKml, ARRAYSIZEOF(szCurrentKml));
			HeapFree(hHeap,0,szTmp);
		}
		#else
		{
			ReadFile(hFile, szCurrentKml, nLength, &lBytesRead, NULL);
		}
		#endif
		if (nLength != lBytesRead) goto read_err;
		szCurrentKml[nLength] = 0;
		break;
	default:
		AbortMessage(_T("This file is for an unknown version of Emu42."));
		goto restore;
	}

	// read chipset size inside file
	ReadFile(hFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesRead, NULL);
	if (lBytesRead != sizeof(lSizeofChipset)) goto read_err;
	if (lSizeofChipset <= sizeof(Chipset))	// actual or older chipset version
	{
		// read chipset content
		ZeroMemory(&Chipset,sizeof(Chipset)); // init chipset
		ReadFile(hFile, &Chipset, lSizeofChipset, &lBytesRead, NULL);
	}
	else									// newer chipset version
	{
		// read my used chipset content
		ReadFile(hFile, &Chipset, sizeof(Chipset), &lBytesRead, NULL);

		// skip rest of chipset
		SetFilePointer(hFile, lSizeofChipset-sizeof(Chipset), NULL, FILE_CURRENT);
		lSizeofChipset = sizeof(Chipset);
	}
	if (lBytesRead != lSizeofChipset) goto read_err;
	Chipset.NCE2 = NULL;
	Chipset.NCE3 = NULL;

	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);

	if (szCurrentKml == NULL)
	{
		if (!DisplayChooseKml(Chipset.type))
			goto restore;
	}
	while (TRUE)
	{
		BOOL bOK;

		bOK = InitKML(szCurrentKml,FALSE);
		bOK = bOK && (cCurrentRomType == Chipset.type);
		if (bOK) break;

		KillKML();
		if (!DisplayChooseKml(Chipset.type))
			goto restore;
	}
	// reload old button state
	ReloadButtons(Chipset.Keyboard_Row,sizeof(Chipset.Keyboard_Row));

	if (Chipset.bSlave)						// 2nd slave controller
	{
		// read slave chipset size inside file
		ReadFile(hFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesRead, NULL);
		if (lBytesRead != sizeof(lSizeofChipset)) goto read_err;

		if (lSizeofChipset <= sizeof(ChipsetS))	// actual or older chipset version
		{
			// read chipset content
			ZeroMemory(&ChipsetS,sizeof(ChipsetS)); // init chipset
			ReadFile(hFile, &ChipsetS, lSizeofChipset, &lBytesRead, NULL);
		}
		else								// newer chipset version
		{
			// read my used chipset content
			ReadFile(hFile, &ChipsetS, sizeof(ChipsetS), &lBytesRead, NULL);

			// skip rest of chipset
			SetFilePointer(hFile, lSizeofChipset-sizeof(ChipsetS), NULL, FILE_CURRENT);
			lSizeofChipset = sizeof(ChipsetS);
		}
		if (lBytesRead != lSizeofChipset) goto read_err;
		ChipsetS.NCE2 = NULL;
		ChipsetS.NCE3 = NULL;
	}

	if (Chipset.NCE2Ram)					// RAM at NCE2 master
	{
		Chipset.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE2Size);
		if (Chipset.NCE2 == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, Chipset.NCE2, Chipset.NCE2Size, &lBytesRead, NULL);
		if (lBytesRead != Chipset.NCE2Size) goto read_err;
	}
	if (Chipset.NCE3Ram)					// RAM at NCE3 master
	{
		Chipset.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE3Size);
		if (Chipset.NCE3 == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, Chipset.NCE3, Chipset.NCE3Size, &lBytesRead, NULL);
		if (lBytesRead != Chipset.NCE3Size) goto read_err;
	}
	if (ChipsetS.NCE2Ram)					// RAM at NCE2 slave
	{
		ChipsetS.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE2Size);
		if (ChipsetS.NCE2 == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, ChipsetS.NCE2, ChipsetS.NCE2Size, &lBytesRead, NULL);
		if (lBytesRead != ChipsetS.NCE2Size) goto read_err;
	}
	if (ChipsetS.NCE3Ram)					// RAM at NCE3 slave
	{
		ChipsetS.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE3Size);
		if (ChipsetS.NCE3 == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, ChipsetS.NCE3, ChipsetS.NCE3Size, &lBytesRead, NULL);
		if (lBytesRead != ChipsetS.NCE3Size) goto read_err;
	}

	LoadBreakpointList(hFile);				// load debugger breakpoint list

	if (Chipset.wRomCrc != CrcRom())		// ROM changed
	{
		CpuReset();
		Chipset.Shutdn = FALSE;				// automatic restart
	}

	lstrcpy(szCurrentFilename, szFilename);
	_ASSERT(hCurrentFile == NULL);
	hCurrentFile = hFile;
	#if defined _USRDLL						// DLL version
		// notify main proc about current document file
		if (pEmuDocumentNotify) pEmuDocumentNotify(szCurrentFilename);
	#endif
	SetWindowTitle(szCurrentFilename);
	UpdateWindowStatus();
	return TRUE;

read_err:
	AbortMessage(_T("This file must be truncated, and cannot be loaded."));
restore:
	if (INVALID_HANDLE_VALUE != hFile)		// close if valid handle
		CloseHandle(hFile);
	RestoreBackup();
	ResetBackup();
	return FALSE;
}

BOOL SaveDocument(VOID)
{
	DWORD           lBytesWritten;
	DWORD           lSizeofChipset;
	UINT            nLength;
	WINDOWPLACEMENT wndpl;

	if (hCurrentFile == NULL) return FALSE;

	_ASSERT(hWnd);							// window open
	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWnd, &wndpl);
	Chipset.nPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPosY = (SWORD) wndpl.rcNormalPosition.top;

	SetFilePointer(hCurrentFile,0,NULL,FILE_BEGIN);

	if (!WriteFile(hCurrentFile, pbySignature, sizeof(pbySignature), &lBytesWritten, NULL))
	{
		AbortMessage(_T("Could not write into file !"));
		return FALSE;
	}

	Chipset.wRomCrc = CrcRom();				// save fingerprint of ROM

	nLength = lstrlen(szCurrentKml);
	WriteFile(hCurrentFile, &nLength, sizeof(nLength), &lBytesWritten, NULL);
	#if defined _UNICODE
	{
		LPSTR szTmp = HeapAlloc(hHeap,0,nLength);
		if (szTmp != NULL)
		{
			WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK,
								szCurrentKml, nLength,
								szTmp, nLength, NULL, NULL);
			WriteFile(hCurrentFile, szTmp, nLength, &lBytesWritten, NULL);
			HeapFree(hHeap,0,szTmp);
		}
	}
	#else
	{
		WriteFile(hCurrentFile, szCurrentKml, nLength, &lBytesWritten, NULL);
	}
	#endif
	lSizeofChipset = sizeof(Chipset);
	WriteFile(hCurrentFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesWritten, NULL);
	WriteFile(hCurrentFile, &Chipset, lSizeofChipset, &lBytesWritten, NULL);
	if (Chipset.bSlave)
	{
		lSizeofChipset = sizeof(ChipsetS);
		WriteFile(hCurrentFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesWritten, NULL);
		WriteFile(hCurrentFile, &ChipsetS, lSizeofChipset, &lBytesWritten, NULL);
	}
	if (Chipset.NCE2Ram)  WriteFile(hCurrentFile, Chipset.NCE2,  Chipset.NCE2Size,  &lBytesWritten, NULL);
	if (Chipset.NCE3Ram)  WriteFile(hCurrentFile, Chipset.NCE3,  Chipset.NCE3Size,  &lBytesWritten, NULL);
	if (ChipsetS.NCE2Ram) WriteFile(hCurrentFile, ChipsetS.NCE2, ChipsetS.NCE2Size, &lBytesWritten, NULL);
	if (ChipsetS.NCE3Ram) WriteFile(hCurrentFile, ChipsetS.NCE3, ChipsetS.NCE3Size, &lBytesWritten, NULL);
	SaveBreakpointList(hCurrentFile);		// save debugger breakpoint list
	SetEndOfFile(hCurrentFile);				// cut the rest
	return TRUE;
}

BOOL SaveDocumentAs(LPCTSTR szFilename)
{
	HANDLE hFile;

	if (hCurrentFile)						// already file in use
	{
		CloseHandle(hCurrentFile);			// close it, even it's same, so data always will be saved
		hCurrentFile = NULL;
	}
	hFile = CreateFile(szFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)		// error, couldn't create a new file
	{
		AbortMessage(_T("This file must be currently used by another instance of Emu42."));
		return FALSE;
	}
	lstrcpy(szCurrentFilename, szFilename);	// save new file name
	hCurrentFile = hFile;					// and the corresponding handle
	#if defined _USRDLL						// DLL version
		// notify main proc about current document file
		if (pEmuDocumentNotify) pEmuDocumentNotify(szCurrentFilename);
	#endif
	SetWindowTitle(szCurrentFilename);		// update window title line
	UpdateWindowStatus();					// and draw it
	return SaveDocument();					// save current content
}



//################
//#
//#    Backup
//#
//################

BOOL SaveBackup(VOID)
{
	WINDOWPLACEMENT wndpl;

	if (pbyRom == NULL) return FALSE;

	// save window position
	_ASSERT(hWnd);							// window open
	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWnd, &wndpl);
	Chipset.nPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPosY = (SWORD) wndpl.rcNormalPosition.top;

	lstrcpy(szBackupFilename, szCurrentFilename);
	lstrcpy(szBackupKml, szCurrentKml);
	if (BackupChipset.NCE2)  HeapFree(hHeap,0,BackupChipset.NCE2);
	if (BackupChipset.NCE3)  HeapFree(hHeap,0,BackupChipset.NCE3);
	if (BackupChipsetS.NCE2) HeapFree(hHeap,0,BackupChipsetS.NCE2);
	if (BackupChipsetS.NCE3) HeapFree(hHeap,0,BackupChipsetS.NCE3);

	CopyMemory(&BackupChipset, &Chipset, sizeof(Chipset));
	if (Chipset.bSlave)						// 2nd slave controller
	{
		CopyMemory(&BackupChipsetS, &ChipsetS, sizeof(ChipsetS));
	}
	if (Chipset.NCE2Ram)					// RAM at NCE2 master
	{
		BackupChipset.NCE2 = HeapAlloc(hHeap,0,Chipset.NCE2Size);
		CopyMemory(BackupChipset.NCE2, Chipset.NCE2, Chipset.NCE2Size);
	}
	if (Chipset.NCE3Ram)					// RAM at NCE3 master
	{
		BackupChipset.NCE3 = HeapAlloc(hHeap,0,Chipset.NCE3Size);
		CopyMemory(BackupChipset.NCE3, Chipset.NCE3, Chipset.NCE3Size);
	}
	if (ChipsetS.NCE2Ram)					// RAM at NCE2 slave
	{
		BackupChipsetS.NCE2 = HeapAlloc(hHeap,0,ChipsetS.NCE2Size);
		CopyMemory(BackupChipsetS.NCE2, ChipsetS.NCE2, ChipsetS.NCE2Size);
	}
	if (ChipsetS.NCE3Ram)					// RAM at NCE3 slave
	{
		BackupChipsetS.NCE3 = HeapAlloc(hHeap,0,ChipsetS.NCE3Size);
		CopyMemory(BackupChipsetS.NCE3, ChipsetS.NCE3, ChipsetS.NCE3Size);
	}
	bBackup = TRUE;
	UpdateWindowStatus();
	return TRUE;
}

BOOL RestoreBackup(VOID)
{
	if (!bBackup) return FALSE;
	ResetDocument();
	// need chipset for contrast setting in InitKML()
	Chipset.contrast = BackupChipset.contrast;
	if (!InitKML(szBackupKml,TRUE))
	{
		InitKML(szCurrentKml,TRUE);
		return FALSE;
	}
	lstrcpy(szCurrentKml, szBackupKml);
	lstrcpy(szCurrentFilename, szBackupFilename);
	if (szCurrentFilename[0])
	{
		hCurrentFile = CreateFile(szCurrentFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if (hCurrentFile == INVALID_HANDLE_VALUE)
		{
			hCurrentFile = NULL;
			szCurrentFilename[0] = 0;
		}
	}
	CopyMemory(&Chipset, &BackupChipset, sizeof(Chipset));
	if (Chipset.bSlave)						// 2nd slave controller
	{
		CopyMemory(&ChipsetS, &BackupChipsetS, sizeof(ChipsetS));
	}
	if (Chipset.NCE2Ram)					// RAM at NCE2 master
	{
		Chipset.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE2Size);
		CopyMemory(Chipset.NCE2, BackupChipset.NCE2, Chipset.NCE2Size);
	}
	if (Chipset.NCE3Ram)					// RAM at NCE3 master
	{
		Chipset.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,Chipset.NCE3Size);
		CopyMemory(Chipset.NCE3, BackupChipset.NCE3, Chipset.NCE3Size);
	}
	if (ChipsetS.NCE2Ram)					// RAM at NCE2 slave
	{
		ChipsetS.NCE2 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE2Size);
		CopyMemory(ChipsetS.NCE2, BackupChipsetS.NCE2, ChipsetS.NCE2Size);
	}
	if (ChipsetS.NCE3Ram)					// RAM at NCE3 slave
	{
		ChipsetS.NCE3 = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,ChipsetS.NCE3Size);
		CopyMemory(ChipsetS.NCE3, BackupChipsetS.NCE3, ChipsetS.NCE3Size);
	}
	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);
	UpdateWindowStatus();
	return TRUE;
}

BOOL ResetBackup(VOID)
{
	if (!bBackup) return FALSE;
	szBackupFilename[0] = 0;
	szBackupKml[0] = 0;
	if (BackupChipset.NCE2)  HeapFree(hHeap,0,BackupChipset.NCE2);
	if (BackupChipset.NCE3)  HeapFree(hHeap,0,BackupChipset.NCE3);
	if (BackupChipsetS.NCE2) HeapFree(hHeap,0,BackupChipsetS.NCE2);
	if (BackupChipsetS.NCE3) HeapFree(hHeap,0,BackupChipsetS.NCE3);
	ZeroMemory(&BackupChipset,sizeof(BackupChipset));
	ZeroMemory(&BackupChipsetS,sizeof(BackupChipsetS));
	bBackup = FALSE;
	UpdateWindowStatus();
	return TRUE;
}



//################
//#
//#    Open File Common Dialog Boxes
//#
//################

static VOID InitializeOFN(LPOPENFILENAME ofn)
{
	ZeroMemory((LPVOID)ofn, sizeof(OPENFILENAME));
	ofn->lStructSize = sizeof(OPENFILENAME);
	ofn->hwndOwner = hWnd;
	ofn->Flags = OFN_EXPLORER|OFN_HIDEREADONLY;
	return;
}

BOOL GetOpenFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter =
		_T("Emu17B   Document (*.E17)\0*.E17\0")
		_T("Emu17BII Document (*.E17)\0*.E17\0")
		_T("Emu19BII Document (*.E19)\0*.E19\0")
		_T("Emu27S   Document (*.E27)\0*.E27\0")
		_T("Emu28S   Document (*.E28)\0*.E28\0")
		_T("Emu42S   Document (*.E42)\0*.E42\0")
		_T("All Files (*.*)\0*.*\0")
		_T("\0\0");
	ofn.nFilterIndex = 7;					// default
	if (cCurrentRomType == 'T')				// Trader, HP17B
	{
		ofn.lpstrDefExt = _T("E17");
		ofn.nFilterIndex = 1;
	}
	if (cCurrentRomType == 'U')				// Trader II, HP17BII
	{
		ofn.lpstrDefExt = _T("E17");
		ofn.nFilterIndex = 2;
	}
	if (cCurrentRomType == 'Y')				// Tycoon II, HP19BII
	{
		ofn.lpstrDefExt = _T("E19");
		ofn.nFilterIndex = 3;
	}
	if (cCurrentRomType == 'M')				// Mentor, HP27S
	{
		ofn.lpstrDefExt = _T("E27");
		ofn.nFilterIndex = 4;
	}
	if (cCurrentRomType == 'O')				// Orlando, HP28S
	{
		ofn.lpstrDefExt = _T("E28");
		ofn.nFilterIndex = 5;
	}
	if (cCurrentRomType == 'D')				// Davinci, HP42S
	{
		ofn.lpstrDefExt = _T("E42");
		ofn.nFilterIndex = 6;
	}
	ofn.lpstrFile = HeapAlloc(hHeap,0,sizeof(szBufferFilename));
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = ARRAYSIZEOF(szBufferFilename);
	ofn.Flags |= OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	if (GetOpenFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap,0,ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap,0,ofn.lpstrFile);
	return TRUE;
}

BOOL GetSaveAsFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter =
		_T("Emu17B   Document (*.E17)\0*.E17\0")
		_T("Emu17BII Document (*.E17)\0*.E17\0")
		_T("Emu19BII Document (*.E19)\0*.E19\0")
		_T("Emu27S   Document (*.E27)\0*.E27\0")
		_T("Emu28S   Document (*.E28)\0*.E28\0")
		_T("Emu42S   Document (*.E42)\0*.E42\0")
		_T("All Files (*.*)\0*.*\0")
		_T("\0\0");
	ofn.nFilterIndex = 7;					// default
	if (cCurrentRomType == 'T')				// Trader, HP17B
	{
		ofn.lpstrDefExt = _T("E17");
		ofn.nFilterIndex = 1;
	}
	if (cCurrentRomType == 'U')				// Trader II, HP17BII
	{
		ofn.lpstrDefExt = _T("E17");
		ofn.nFilterIndex = 2;
	}
	if (cCurrentRomType == 'Y')				// Tycoon II, HP19BII
	{
		ofn.lpstrDefExt = _T("E19");
		ofn.nFilterIndex = 3;
	}
	if (cCurrentRomType == 'M')				// Mentor, HP27S
	{
		ofn.lpstrDefExt = _T("E27");
		ofn.nFilterIndex = 4;
	}
	if (cCurrentRomType == 'O')				// Orlando, HP28S
	{
		ofn.lpstrDefExt = _T("E28");
		ofn.nFilterIndex = 5;
	}
	if (cCurrentRomType == 'D')				// Davinci, HP42S
	{
		ofn.lpstrDefExt = _T("E42");
		ofn.nFilterIndex = 6;
	}
	ofn.lpstrFile = HeapAlloc(hHeap,0,sizeof(szBufferFilename));
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = ARRAYSIZEOF(szBufferFilename);
	ofn.Flags |= OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
	if (GetSaveFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap,0,ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap,0,ofn.lpstrFile);
	return TRUE;
}

BOOL GetLoadObjectFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	if (cCurrentRomType == 'D')				// Davinci, HP42S
	{
		ofn.lpstrFilter =
			_T("User Code Files (*.RAW)\0*.RAW\0")
			_T("All Files (*.*)\0*.*\0")
			_T("\0\0");
		ofn.lpstrDefExt = _T("RAW");
	}
	else
	{
		ofn.lpstrFilter =
			_T("All Files (*.*)\0*.*\0")
			_T("\0\0");
	}
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap,0,sizeof(szBufferFilename));
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = ARRAYSIZEOF(szBufferFilename);
	ofn.Flags |= OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	if (GetOpenFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap,0,ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap,0,ofn.lpstrFile);
	return TRUE;
}

BOOL GetSaveObjectFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	if (cCurrentRomType == 'D')				// Davinci, HP42S
	{
		ofn.lpstrFilter =
			_T("User Code Files (*.RAW)\0*.RAW\0")
			_T("All Files (*.*)\0*.*\0")
			_T("\0\0");
		ofn.lpstrDefExt = _T("RAW");
	}
	else
	{
		ofn.lpstrFilter =
			_T("All Files (*.*)\0*.*\0")
			_T("\0\0");
	}
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap,0,sizeof(szBufferFilename));
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = ARRAYSIZEOF(szBufferFilename);
	ofn.Flags |= OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
	if (GetSaveFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap,0,ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap,0,ofn.lpstrFile);
	return TRUE;
}



//################
//#
//#    Load and Save HP28S Objects
//#
//################

BOOL LoadObject(LPCTSTR szFilename)			// separated stack writing part
{
	HANDLE hFile;
	DWORD  dwFileSizeLow;
	DWORD  dwFileSizeHigh;
	LPBYTE lpBuf;
	WORD   wError;

	hFile = CreateFile(szFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return FALSE;
	dwFileSizeLow = GetFileSize(hFile, &dwFileSizeHigh);
	if (dwFileSizeHigh != 0)
	{ // file is too large.
		CloseHandle(hFile);
		return FALSE;
	}
	lpBuf = HeapAlloc(hHeap,0,dwFileSizeLow*2);
	if (lpBuf == NULL)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	ReadFile(hFile, lpBuf+dwFileSizeLow, dwFileSizeLow, &dwFileSizeHigh, NULL);
	CloseHandle(hFile);

	wError = WriteStack(lpBuf,dwFileSizeLow);

	if (wError == S_ERR_BINARY)
		AbortMessage(_T("The calculator does not have enough\nfree memory to load this binary file."));

	if (wError == S_ERR_ASCII)
		AbortMessage(_T("The calculator does not have enough\nfree memory to load this text file."));

	HeapFree(hHeap,0,lpBuf);
	return (wError == S_ERR_NO);
}

BOOL SaveObject(LPCTSTR szFilename)			// separated stack reading part
{
	HANDLE	hFile;
	DWORD	lBytesWritten;
	DWORD   dwAddress;
	INT     nLength;

	dwAddress = RPL_Pick(1);
	if (dwAddress == 0)
	{
		AbortMessage(_T("Too Few Arguments."));
		return FALSE;
	}

	hFile = CreateFile(szFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		AbortMessage(_T("Cannot open file."));
		return FALSE;
	}

	WriteFile(hFile, BINARYHEADER28, 8, &lBytesWritten, NULL);

	nLength = RPL_SkipOb(dwAddress) - dwAddress;

	for (; nLength > 0; nLength -= 2)
	{
		BYTE byByte = Read2(dwAddress);
		if (nLength == 1) byByte &= 0xF;
		WriteFile(hFile, &byByte, sizeof(byByte), &lBytesWritten, NULL);
		dwAddress += 2;
	}
	CloseHandle(hFile);
	return TRUE;
}



//################
//#
//#    Load Bitmap
//#
//################

static __inline UINT DibNumColors(BITMAPINFOHEADER CONST *lpbi)
{
	if (lpbi->biClrUsed != 0) return (UINT)lpbi->biClrUsed;

	/* a 24 bitcount DIB has no color table */
	return (lpbi->biBitCount <= 8) ? (1 << lpbi->biBitCount) : 0;
}

static HPALETTE CreateBIPalette(BITMAPINFOHEADER CONST *lpbi)
{
	LOGPALETTE* pPal;
	HPALETTE    hpal = NULL;
	UINT        nNumColors;
	BYTE        red;
	BYTE        green;
	BYTE        blue;
	UINT        i;
	RGBQUAD*    pRgb;

	if (!lpbi)
		return NULL;

	if (lpbi->biSize != sizeof(BITMAPINFOHEADER))
		return NULL;

	// Get a pointer to the color table and the number of colors in it
	pRgb = (RGBQUAD FAR *)((LPBYTE)lpbi + (WORD)lpbi->biSize);
	nNumColors = DibNumColors(lpbi);

	if (nNumColors)
	{
		// Allocate for the logical palette structure
		pPal = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
		if (!pPal)
			return NULL;

		pPal->palNumEntries = nNumColors;
		pPal->palVersion    = 0x300;

		// Fill in the palette entries from the DIB color table and
		// create a logical color palette.
		for (i = 0; i < nNumColors; i++)
		{
			pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
			pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
			pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
			pPal->palPalEntry[i].peFlags = (BYTE)0;
		}
		hpal = CreatePalette(pPal);
		HeapFree(hHeap,0,pPal);
	}
	else
	{
		if (lpbi->biBitCount == 24)
		{
			// A 24 bitcount DIB has no color table entries so, set the
			// number of to the maximum value (256).
			nNumColors = 256;
			pPal = HeapAlloc(hHeap,HEAP_ZERO_MEMORY,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
			if (!pPal)
				return NULL;

			pPal->palNumEntries = nNumColors;
			pPal->palVersion    = 0x300;

			red = green = blue = 0;

			// Generate 256 (= 8*8*4) RGB combinations to fill the palette
			// entries.
			for (i = 0; i < pPal->palNumEntries; i++)
			{
				pPal->palPalEntry[i].peRed   = red;
				pPal->palPalEntry[i].peGreen = green;
				pPal->palPalEntry[i].peBlue  = blue;
				pPal->palPalEntry[i].peFlags = (BYTE)0;

				if (!(red += 32))
					if (!(green += 32))
					blue += 64;
			}
			hpal = CreatePalette(pPal);
			HeapFree(hHeap,0,pPal);
		}
	}
	return hpal;
}

HBITMAP LoadBitmapFile(LPCTSTR szFilename)
{
	HANDLE  hFile;
	HANDLE  hMap;
	LPBYTE  pbyFile;
	HBITMAP hBitmap;
	LPBITMAPFILEHEADER pBmfh;
	LPBITMAPINFO pBmi;

	if (pbyRom == NULL) return NULL;
	hFile = CreateFile(FullFilename(szFilename), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (hMap == NULL)
	{
		CloseHandle(hFile);
		return NULL;
	}
	pbyFile = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
	if (pbyFile == NULL)
	{
		CloseHandle(hMap);
		CloseHandle(hFile);
		return NULL;
	}

	hBitmap = NULL;
	pBmfh = (LPBITMAPFILEHEADER)pbyFile;
	if (pBmfh->bfType != 0x4D42) goto quit; // "BM"
	pBmi = (LPBITMAPINFO)(pbyFile+sizeof(BITMAPFILEHEADER));

	_ASSERT(hPalette == NULL);				// resource free
	hPalette = CreateBIPalette(&pBmi->bmiHeader);
	// save old palette
	hOldPalette = SelectPalette(hWindowDC, hPalette, FALSE);
	RealizePalette(hWindowDC);

	hBitmap = CreateDIBitmap(
		hWindowDC,
		&pBmi->bmiHeader,
		CBM_INIT,
		pbyFile + pBmfh->bfOffBits,
		pBmi, DIB_RGB_COLORS);
	_ASSERT(hBitmap != NULL);

quit:
	UnmapViewOfFile(pbyFile);
	CloseHandle(hMap);
	CloseHandle(hFile);
	return hBitmap;
}
